// Average a MeasurementSet over time.
//
// Written by Mark Kettenis <kettenis@jive.nl>.
//

#define RCSID(string) \
  static char rcsid[] __attribute__ ((unused)) = string

RCSID("@(#) $Id: tmav.cc,v 1.1.2.3 2005/09/02 10:23:10 kettenis Exp $");

#include <aips/aips.h>

#ifdef CASA_AIPS_H

#include <casa/Arrays.h>
#include <casa/Utilities.h>
#include <measures/Measures.h>
#include <measures/Measures/MBaseline.h>
#include <measures/Measures/MDirection.h>
#include <measures/Measures/MEpoch.h>
#include <measures/Measures/MPosition.h>
#include <measures/Measures/Muvw.h>
#include <ms/MeasurementSets.h>
#include <scimath/Mathematics.h>
#include <tables/Tables.h>
#include <tables/Tables/TableIter.h>

using namespace casa;

#else  // not CASA_AIPS_H

#include <aips/Arrays.h>
#include <aips/Mathematics.h>
#include <aips/Measures.h>
#include <aips/Measures/MBaseline.h>
#include <aips/Measures/MDirection.h>
#include <aips/Measures/MEpoch.h>
#include <aips/Measures/MPosition.h>
#include <aips/Measures/Muvw.h>
#include <aips/MeasurementSets.h>
#include <aips/Tables.h>
#include <aips/Tables/TableIter.h>
#include <aips/Utilities.h>

#endif // CASA_AIPS_H

#include <cassert>
#include <cstdlib>
#include <iostream>
#include <map>

// Define if you want better UVW's (see below).
#undef CORRECT_UVWS

// Key type.
struct baselineKey
{
  Int antenna1, antenna2;
};

// Key comparison functor.
struct baselineKeyCompare
{
  // Return true if K is less than L.
  //
  bool operator()(baselineKey k, baselineKey l) const
    {
      return ((k.antenna1 == l.antenna1)
	      ? k.antenna2 < l.antenna2
	      : k.antenna1 < l.antenna1);
    }
};

typedef std::map<baselineKey, MBaseline, baselineKeyCompare> baselineMap;
typedef std::map<Int, MDirection> fieldMap;

class UVWCalculator
{
public:
  UVWCalculator(const MeasurementSet& ms);

  Muvw calcUVW(Int antenna1, Int antenna2, Int fieldID, MEpoch& time);

private:
  const MeasurementSet& ms;

  mutable baselineMap baselineCache;
  const MBaseline& getBaseline(Int antenna1, Int antenna2);

  mutable fieldMap directionCache;
  const MDirection& getDirection(Int fieldId);

#ifdef CORRECT_UVWS
  mutable MeasFrame frame;
  mutable MBaseline::Ref ref;
  mutable MBaseline::Convert conv;
#else
  MEpoch::Convert conv;
  mutable MVEpoch timeCache;
  mutable MEpoch gast;
#endif
};

UVWCalculator::UVWCalculator(const MeasurementSet& ms)
  : ms(ms)
{
#ifdef CORRECT_UVWS
  static const MPosition pos;

  frame = MeasFrame(pos);
  ref = MBaseline::Ref(MBaseline::J2000, frame);
  conv = MBaseline::Convert(MBaseline::ITRF, ref);
#else
  conv = MEpoch::Convert(MEpoch::UTC, MEpoch::GAST);
#endif
}

const MBaseline&
UVWCalculator::getBaseline(Int antenna1, Int antenna2)
{
  struct baselineKey key = { antenna1, antenna2 };

  if (baselineCache.count(key) > 0)
    return baselineCache[key];

  ROMSAntennaColumns antennaColumns(ms.antenna());

  // Construct the baseline based on the antenna positions.
  //
  MPosition pos1(antennaColumns.positionMeas()(antenna1));
  MPosition pos2(antennaColumns.positionMeas()(antenna2));
  assert(pos1.getRef().getType() == MPosition::ITRF);
  assert(pos2.getRef().getType() == MPosition::ITRF);
  MBaseline base(MVBaseline(pos2.getValue(), pos1.getValue()));
  assert(base.getRef().getType() == MBaseline::ITRF);

  return (baselineCache[key] = base);
}

const MDirection&
UVWCalculator::getDirection(Int fieldId)
{
  if (directionCache.count(fieldId) > 0)
    return directionCache[fieldId];

  ROMSFieldColumns fieldColumns(ms.field());

  MDirection dir(fieldColumns.phaseDirMeas(fieldId));
  assert(dir.getRef().getType() == MDirection::J2000);

  return (directionCache[fieldId] = dir);
}

// Calculate the UVW for the baseline ANTENNA1 to ANTENNA2 in the
// direction specified by FIELDID at TIME.
//
// Currently this uses Harro's broken algorithm.
//
Muvw
UVWCalculator::calcUVW(Int antenna1, Int antenna2, Int fieldId, MEpoch& time)
{
  MBaseline base(getBaseline(antenna1, antenna2));
  MDirection dir(getDirection(fieldId));

#ifdef CORRECT_UVWS
  // This is the correct way to calculate the UVW coordinates, at
  // least for the J2000 frame.  Presumably that's what we want in the
  // FITS-IDI file.  It differs from the AIPS++ default for UVWs,
  // which is ITRF.  Unfortunately, calculating the UVWs properly is
  // quite a bit more expensive, so in the end we might not want to do
  // this, and simply average the UVWs instead.
  //
  frame.set(time);
  frame.set(dir);

  // Incorrectly labelled as ITRF.
  return Muvw(MVuvw(conv(base).getValue(), dir.getValue()));
#else
  // The incorrect way.  This doesn't fully take into account
  // precession, nutation, etc.  So it gets worse as time progresses.
  // It's fast though, so we recalculate the UVWs instead of simply
  // averaging them.
  //
  // Convert the time to GAST.  Speedup the common case of conversion
  // from UTC by using a static convertor.
  //
  assert (time.getRef().getType() == MEpoch::UTC);
  if (time.getValue() != timeCache)
    {
      timeCache = time.getValue();
      gast = conv(time);
    }

  // Use the converted time to offset the hour angle.
  dir.shiftLongitude(-gast.getValue().getDayFraction() * 2 * C::pi);

  // Incorrectly labelled as ITRF.
  return Muvw(MVuvw(base.getValue(), dir.getValue()));
#endif
}

// Compare row ROW1 from MeasurementSet MS1 with ROW2 from
// MeasurementSet MS2.  Return true if those tow can be averaged over,
// false otherwise.

static Bool
keyCompare(MeasurementSet& ms1, uInt row1, MeasurementSet& ms2, uInt row2)
{
  Block<String> columns(11);
  columns[0] = "ANTENNA1";
  columns[1] = "ANTENNA2";
  columns[2] = "FEED1";
  columns[3] = "FEED2";
  columns[4] = "DATA_DESC_ID";
  columns[5] = "PROCESSOR_ID";
  columns[6] = "FIELD_ID";
  columns[7] = "SCAN_NUMBER";
  columns[8] = "ARRAY_ID";
  columns[9] = "OBSERVATION_ID";
  columns[10] = "STATE_ID";

  for (uInt i = 0; i < columns.nelements(); i++)
    {
      ROScalarColumn<Int> c1(ms1, columns[i]);
      ROScalarColumn<Int> c2(ms2, columns[i]);
      if (c1(row1) != c2(row2))
	return False;
    }

  return True;
}

int
main(int argc, char **argv)
{
  bool usage = true;
  uInt reductionFactor = 0;
  int c;

  while ((c = ::getopt(argc, argv, "c:r:")) != -1)
    {
      switch (c)
	{
	case 'r':
	  reductionFactor = std::atoi(optarg);
	  usage = false;
	  break;
	case '?':
	  usage = true;
	  break;
	}
    }

  if (usage || optind != (argc - 2))
    {
      std::cerr << "usage: " << argv[0]
		<< " -r REDUCTION source target" << endl;
      return EXIT_FAILURE;
    }

  if (reductionFactor < 1)
    {
      std::cerr << "reduction factor should at least be 1" << endl;
      return EXIT_FAILURE;
    }

  try
    {
      MeasurementSet src(argv[optind++]);
      ROMSMainColumns srcColumns(src);
      UVWCalculator calc(src);

      // Create a new MeasurementSet.  Copy the table info and all
      // subtables verbatim; time averaging should not have any effect
      // on them.
      //
      TableDesc tabDesc = src.tableDesc();
      SetupNewTable setup(argv[optind], tabDesc, Table::New);
      MeasurementSet tgt(setup);
      MSMainColumns tgtColumns(tgt);
      TableCopy::copyInfo(tgt, src);
      TableCopy::copySubTables(tgt, src);

      Block<String> iterKeys(1);
      iterKeys[0] = "TIME";
      TableIterator iter(src, iterKeys, TableIterator::Ascending);

      uInt iteration = 0;
      while(!iter.pastEnd())
	{
	  MeasurementSet src(iter.table());

	  if (iteration == 0)
	    {
	      TableCopy::copyRows(tgt, src, tgt.nrow(), 0, src.nrow());
	    }
	  else
	    {
	      ROMSMainColumns srcColumns(src);

	      // We assume that for each time interval, we have the
	      // same number of interferometers, and that they're
	      // ordered in the same fashion.  However, if that's not
	      // the case, the program won't fall over; it just won't
	      // average.
	      for (uInt srcRow = 0; srcRow < src.nrow(); srcRow++)
		{
		  uInt tgtRow = tgt.nrow() - src.nrow() + srcRow;
		  if (!keyCompare(tgt, tgtRow, src, srcRow))
		    {
		      iteration = 0;
		      break;
		    }

		  // Only average if we have consequetive time
		  // intervals.
		  Double endTime = tgtColumns.time()(tgtRow)
		    + tgtColumns.interval()(tgtRow) / 2;
		  Double startTime = srcColumns.time()(srcRow)
		    - srcColumns.interval()(srcRow) / 2;
		  if (!nearAbs(endTime, startTime, 1.0e-3))
		    {
		      std::cerr << "Gap!" << endl;
		      iteration = 0;
		      break;
		    }
		}

	      if (iteration == 0)
		continue;

	      Float alpha = 1.0 / (iteration + 1);
	      Float beta = 1 - alpha;

	      for (uInt srcRow = 0; srcRow < src.nrow(); srcRow++)
		{
		  uInt tgtRow = tgt.nrow() - src.nrow() + srcRow;

		  Matrix<Complex> srcData = srcColumns.data()(srcRow);
		  Matrix<Complex> tgtData = tgtColumns.data()(tgtRow);
		  Matrix<Complex> newData(srcData.shape());
		  Vector<Float> srcWeight = srcColumns.weight()(srcRow);
		  Vector<Float> tgtWeight = tgtColumns.weight()(tgtRow);
		  Vector<Float> newWeight;
		  Quantity srcInterval = srcColumns.intervalQuant()(srcRow);
		  Quantity tgtInterval = tgtColumns.intervalQuant()(tgtRow);
		  Quantity newInterval = srcInterval + tgtInterval;
		  MEpoch tgtTime = tgtColumns.timeMeas()(tgtRow);
		  MEpoch newTime(tgtTime.getValue() + srcInterval / 2,
				 tgtTime.getRef());

		  // Treat negative weights as being zero, effectively
		  // ommitting the data for this row, unless we're
		  // averaging over rows that both have a negative
		  // weights.
		  assert (srcWeight.nelements() == tgtWeight.nelements());
		  for (uInt corr = 0; corr < srcWeight.nelements(); corr++)
		    {
		      if (srcWeight(corr) < 0 && tgtWeight(corr) > 0)
			srcWeight(corr) = 0;
		      if (tgtWeight(corr) < 0 && srcWeight(corr) > 0)
			tgtWeight(corr) = 0;
		    }

		  newWeight = beta * tgtWeight + alpha * srcWeight;
		  for (uInt corr = 0; corr < srcData.nrow(); corr++)
		    {
		      Float alphaWeight, betaWeight;

		      alphaWeight = fabs(srcWeight(corr) / newWeight(corr));
		      betaWeight = fabs(tgtWeight(corr) / newWeight(corr));

		      for (uInt chan = 0; chan < srcData.ncolumn(); chan++)
			{
			  newData(corr, chan) =
			    (betaWeight * tgtData(corr, chan));
			  newData(corr, chan) +=
			    (alphaWeight * srcData(corr, chan));
			}
		    }
		  tgtColumns.weight().put(tgtRow, newWeight);
		  tgtColumns.data().put(tgtRow, newData);
		  tgtColumns.intervalQuant().put(tgtRow, newInterval);
		  tgtColumns.timeMeas().put(tgtRow, newTime);

		  // A bit crude, but at least it's correct.
		  if (srcColumns.flagRow()(srcRow))
		    tgtColumns.flagRow().put(tgtRow, True);
		}
	    }

	  ++iteration %= reductionFactor;
	  iter.next();
	}

      // Recalculate the UVW's.
      for (uInt row; row < tgtColumns.nrow(); row++)
	{
	  uInt antenna1 = tgtColumns.antenna1()(row);
	  uInt antenna2 = tgtColumns.antenna2()(row);
	  uInt fieldId = tgtColumns.fieldId()(row);
	  MEpoch time = tgtColumns.timeMeas()(row);

	  Muvw newUVW = calc.calcUVW(antenna1, antenna2, fieldId, time);

	  tgtColumns.uvwMeas().put(row, newUVW);
	}

      // What should we do with the FLAG and FLAG_CATEGORY columns?
      // We should probably just clear them, and re-popagate the
      // flagging information.  For now, we don't do anything.
    }

  catch (AipsError error)
    {
      std::cerr << error.getMesg() << endl;
      return EXIT_FAILURE;
    }

  return EXIT_SUCCESS;
}
