#include <jive/j2ms/SFXCVisBuffer.h>
#include <jive/j2ms/Experiment.h>
#include <jive/j2ms/COFScan.h>
#include <jive/j2ms/Scan.h>
#include <jive/j2ms/DayConversion.h>
#include <jive/DtdLUT/CorrelationCode.h>
#include <jive/Utilities/jexcept.h>

#include <casa/Quanta/MVTime.h>
#include <casa/Utilities/Regex.h>

#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <libgen.h>
#include <limits.h>
#include <stdlib.h>

#include <cmath>
#include <list>
#include <string>
#include <algorithm>

#include <jive/j2ms/sfxc/output_header.h>


SFXCVisBuffer::SFXCVisBuffer(const casa::String &vb,
			     const Experiment *exp,
			     const string& /*vbo*/) :
  VisibilityBuffer(exp),
  __rootdir(exp->getRootDir() + "/" + vb)
{
    // Transform the experiment's scanlist into a list of scanpointers
    const j2ms::scaninfolist_t& scanlist(exp->getScanlist());
    j2ms::scaninfolist_t::const_iterator scan;
    
    for (scan = scanlist.begin(); scan != scanlist.end(); scan++)
      __scanptrlist.push_back(new j2ms::COFScan(this, *scan));

    // Construct a list of all scans to fool the Ghostbuster filter.
    for (scan = scanlist.begin(); scan != scanlist.end(); scan++)
      stationscanlist.push_back(j2ms::stationscaninfo_t((*scan).scanStart, (*scan).scanEnd));

    char buf[PATH_MAX];
    if (::realpath(__rootdir.c_str(), buf) == NULL)
      THROW_JEXCEPT("Unresolvable path" << __rootdir);

    const char *name = ::basename(buf);
    if (name == NULL)
      THROW_JEXCEPT("Cannot get base portion of directory name " << __rootdir);

    string path = __rootdir + "/" + name + ".cor";
    fd = ::open(path.c_str(), O_RDONLY | O_LARGEFILE);
    if (fd == -1)
      THROW_JEXCEPT("Cannot open " << path);

    struct Output_header_global hdr;
    ssize_t nbytes = ::read(fd, &hdr, sizeof hdr);
    if (nbytes != sizeof hdr)
      THROW_JEXCEPT("Cannot read global haeder");

    if (hdr.header_size != sizeof (hdr))
      THROW_JEXCEPT("Global header invalid");

    integration_time = std::pow(2.0, hdr.integration_time);

    unsigned month, day;
    DayConversion::dayNrToMonthDay(month, day, hdr.start_day, hdr.start_year);
    double dayfraction = (double) hdr.start_time / DayConversion::secondsPerDay;
    casa::MVTime start = casa::MVTime(hdr.start_year, month + 1, day, dayfraction);
    start_time = casa::MVEpoch(start) + casa::Quantity(integration_time / 2, "s");

    mepoch_in_scaninfo cond(start_time);
    scanptrlist_type::const_iterator scanptr;
    scanptr = find_scaninfo_by_time(__scanptrlist, cond);
    __mode = (*scanptr)->getFrequencyCode();

    createAntennaMapping();
    createSubbandMapping();

    baseline = baselines.end();
    subband = subbands.end();

    num_channels = hdr.number_channels;

    correlationset_t pols;
    switch (hdr.polarisation_type)
      {
      case Output_header_global::LEFT_POLARISATION:
	pols.insert(CorrelationCode(j2ms::l, j2ms::l));
	break;
      case Output_header_global::RIGHT_POLARISATION:
	pols.insert(CorrelationCode(j2ms::r, j2ms::r));
	break;
      case Output_header_global::LEFT_RIGHT_POLARISATION:
	pols.insert(CorrelationCode(j2ms::r, j2ms::r));
	pols.insert(CorrelationCode(j2ms::l, j2ms::l));
	break;
      case Output_header_global::LEFT_RIGHT_POLARISATION_WITH_CROSSES:
	pols.insert(CorrelationCode(j2ms::r, j2ms::r));
	pols.insert(CorrelationCode(j2ms::r, j2ms::l));
	pols.insert(CorrelationCode(j2ms::l, j2ms::r));
	pols.insert(CorrelationCode(j2ms::l, j2ms::l));
	break;
      default:
	THROW_JEXCEPT("Unknown polarisations in header");
    }

    j2ms::log2polmap l2p;
    switch (hdr.polarisation_type)
      {
      case Output_header_global::LEFT_POLARISATION:
	l2p.insert(std::make_pair(0, j2ms::l));
	break;
      case Output_header_global::RIGHT_POLARISATION:
	l2p.insert(std::make_pair(0, j2ms::r));
	break;
      default:
	l2p.insert(std::make_pair(0, j2ms::r));
	l2p.insert(std::make_pair(1, j2ms::l));
	break;
      }

    __setup = CorrelatorSetup(num_channels, pols, integration_time,
			      CorrelatorSetup::bypassVersion, l2p, -1);

    // Now we can pre-allocate the temporary storage.
    cplxbuf.resize(__setup.nrOfSpectralPoints());
    weight.resize(__setup.getCorrelations().size());
    data.resize(__setup.getCorrelations().size(),  __setup.nrOfSpectralPoints());
    cplxnullen = complexvector_type(__setup.nrOfSpectralPoints(), casa::Complex(0.0f, 0.0f));

    // Move forward to first visibility.
    next();
}

void
SFXCVisBuffer::createAntennaMapping(void)
{
  const Experiment& exp(getExperiment());
  const AntennaDB& adb(exp.getAntennaDB());
  std::vector<casa::String> antennas;
  std::vector<casa::String>::const_iterator antenna;
  AntennaInfo ai;

  // Convert from casa::Vector to std::vector.
  exp.getAntennas().tovector(antennas);

  // Sort antennas in alphabetical order.
  std::sort(antennas.begin(), antennas.end());

  for (antenna = antennas.begin(); antenna != antennas.end(); antenna++)
    {
      adb.searchAntenna(ai, *antenna);
      antenna_map.push_back(ai.getId());
    }
}

void
SFXCVisBuffer::createSubbandMapping(void)
{
  const Experiment& exp(getExperiment());
  const FrequencyDB& fdb(exp.getFrequencyDB());
  FrequencyConfig fc;

  fdb.searchFrequencyConfig(fc, __mode);
  const basebandlist_t& bbs(fc.getBaseBands());
  basebandlist_t::const_iterator bb;

  // Figure out whether this is a dual sideband experiment or not.
  bool dual_sideband = false;
  for (bb = bbs.begin(); bb != bbs.end(); bb++)
    {
      if ((*bb).getSideBand() != (*bbs.begin()).getSideBand())
	{
	  dual_sideband = true;
	  break;
	}
    }

  // Sort frequencies.
  std::set<casa::Double> freqs;
  for (bb = bbs.begin(); bb != bbs.end(); bb++)
    freqs.insert((*bb).getStartfreq());

  // And build a mapping based on the ordered frequencies.
  std::set<casa::Double>::const_iterator freq;
  int frequency_nr = 0;
  for (freq = freqs.begin(); freq != freqs.end(); freq++)
    {
      for (bb = bbs.begin(); bb != bbs.end(); bb++)
	{
	  if ((*bb).getStartfreq() == *freq)
	    {
	      int sideband;

	      if ((*bb).getSideBand() == BaseBand::lower)
		sideband = 0;
	      else
		sideband = 1;

	      std::pair<int, int> sb = std::make_pair(frequency_nr, sideband);
	      subband_map[sb] = (*bb).getBaseBandNr();

	      if (!dual_sideband || (*bb).getSideBand() == BaseBand::upper)
		frequency_nr++;
	    }
	}
    }
}

casa::Bool
SFXCVisBuffer::getCurrentVisibility(Visibility &v)
{
  if (subbands.empty() || baseline == baselines.end())
    return casa::False;

  if ((*baseline).first >= antenna_map.size()
      || (*baseline).second >= antenna_map.size())
    THROW_JEXCEPT("Unobserved station");
  if (subband_map.count(*subband) == 0)
    THROW_JEXCEPT("Unobserved subband");

  const correlationset_t &pols(__setup.getCorrelations());
  correlationset_t::const_iterator pol;

  bool swap = true;
  casa::Int ant1 = antenna_map[(*baseline).first];
  casa::Int ant2 = antenna_map[(*baseline).second];
  if (ant1 > ant2)
    {
      swap = false;
      ant1 = antenna_map[(*baseline).second];
      ant2 = antenna_map[(*baseline).first];
    }

  int i;
  for (pol = pols.begin(), i = 0; pol != pols.end(); pol++, i++)
    {
      std::vector<SFXCBaselineKey>::const_iterator key;

      int j;
      for (key = keys.begin(), j = 0; key != keys.end(); key++, j++)
	{
	  if ((*key).baseline != (*baseline) || (*key).subband != (*subband))
	    continue;

	  std::pair<int, int> polarisation = std::make_pair((*pol)[0] == j2ms::l, (*pol)[1] == j2ms::l);
	  if ((*key).polarisation == polarisation)
	    break;
	}

      if (key == keys.end())
	{
	  data.row(i) = cplxnullen;
	  weight(i) = 0.0;
	  continue;
	}

      if ((*key).subband.second == 0)
	for (int channel = 0; channel < num_channels; channel++)
	  cplxbuf[num_channels - channel - 1] = casa::Complex(visibilities[j][channel * 2], visibilities[j][channel * 2 + 1]);
      else
	for (int channel = 0; channel < num_channels; channel++)
	  cplxbuf[channel] = casa::Complex(visibilities[j][channel * 2], visibilities[j][channel * 2 + 1]);

      if (swap)
	{
	  for (int channel = 0; channel < num_channels; channel++)
	    cplxbuf[channel] = std::conj(cplxbuf[channel]);
	}

      data.row(i) = cplxbuf;
      weight(i) = 1.0;
    }

  ExtVisLabel lab;

  lab.correlations.clear();
  for (pol = pols.begin(); pol != pols.end(); pol++)
    lab.correlations.push_back(*pol);
  lab.subbandnr = subband_map[*subband];
  lab.baseline = BaseLineIdx(ant1, ant2);

  if (uvw_map.count(ant1) == 0 || uvw_map.count(ant2) == 0)
    THROW_JEXCEPT("Missing uvw");

  casa::Muvw::Ref ref(casa::Muvw::J2000);
  casa::Muvw uvw = casa::Muvw(uvw_map[ant1] - uvw_map[ant2], ref);
  v = Visibility(lab, data, Visibility::frequency, casa::False, weight, slice_time, uvw, this);

  return casa::True;
}

void
SFXCVisBuffer::read_timeslice_header(void)
{
  struct Output_header_timeslice hdr;
  ssize_t nbytes = ::read(fd, &hdr, sizeof hdr);
  if (nbytes != sizeof hdr)
    {
      if (nbytes == 0)
	{
	  next_slice = -1;
	  return;
	}
      THROW_JEXCEPT("Cannot read timeslice header");
    }

  next_slice = hdr.integration_slice;
  next_baselines = hdr.number_baselines;

  uvw_map.clear();
  for (int i = 0; i < hdr.number_uvw_coordinates; i++)
    {
      struct Output_uvw_coordinates uvw;
      ssize_t nbytes = ::read(fd, &uvw, sizeof uvw);
      if (nbytes != sizeof uvw)
	THROW_JEXCEPT("Cannot read uvw");

      if ((unsigned int)uvw.station_nr >= antenna_map.size())
	THROW_JEXCEPT("Unobserved station");
      uvw_map[antenna_map[uvw.station_nr]] = casa::MVuvw(uvw.u, uvw.v, uvw.w);
    }
}

void
SFXCVisBuffer::read_baseline(void)
{
  struct Output_header_baseline hdr;
  ssize_t nbytes = ::read(fd, &hdr, sizeof hdr);
  if (nbytes != sizeof hdr)
    THROW_JEXCEPT("Cannot read baseline header");

  if (current_baseline >= visibilities.size())
      visibilities.push_back(new float[(num_channels + 1) * 2]);

  nbytes = ::read(fd, visibilities[current_baseline], (num_channels + 1) * 2 * sizeof(float));
  if (nbytes != (ssize_t)((num_channels + 1) * 2 * sizeof(float)))
    THROW_JEXCEPT("Cannot read baseline visibilities");

  SFXCBaselineKey key;
  key.baseline = std::make_pair(hdr.station_nr1, hdr.station_nr2);
  key.polarisation = std::make_pair(hdr.polarisation1, hdr.polarisation2);
  key.subband = std::make_pair(hdr.frequency_nr, hdr.sideband);
  keys.push_back(key);

  baselines.insert(key.baseline);
  subbands.insert(key.subband);

  baseline = baselines.begin();
  subband = subbands.begin();

  current_baseline++;
}

void
SFXCVisBuffer::read_timeslice(void)
{
  current_baseline = 0;

  do {
    while (next_baselines--)
      read_baseline();
    read_timeslice_header();
  } while (current_slice == next_slice);
}

casa::Bool
SFXCVisBuffer::next(void)
{
  if (subbands.empty())
    {
      read_timeslice_header();
    }
  else
    {
      subband++;
      if (subband == subbands.end())
	{
	  baseline++;
	  subband = subbands.begin();
	}
    }

  if (baseline == baselines.end())
    {
      keys.clear();
      baselines.clear();
      subbands.clear();

      if (next_slice == -1)
	return casa::False;

      current_slice = next_slice;
      slice_time = start_time +
	casa::Quantity(current_slice * integration_time, "s");
      read_timeslice();

      mepoch_in_scaninfo cond(slice_time);
      scanptrlist_type::const_iterator scan;
      scan = find_scaninfo_by_time(__scanptrlist, cond);
      if (scan != __scanptrlist.end())
	  setCurrentScanptr(*scan);
    }

  return casa::True;
}

const casa::String&
SFXCVisBuffer::getRootDir() const
{
  return __rootdir;
}

casa::Int
SFXCVisBuffer::getVisBufId() const
{
  return 0;
}

CorrelatorSetup
SFXCVisBuffer::correlatorSetup() const
{
  return __setup;
}

const j2ms::stationscanlist_t&
SFXCVisBuffer::getStationScanlist(int) const
{
  // Simply return a list with all scans in the experiment.  This
  // should disable the Ghostbuster filter completely.
  return stationscanlist;
}

SFXCVisBuffer::~SFXCVisBuffer()
{
  scanptrlist_type::iterator scan;

  for(scan = __scanptrlist.begin(); scan != __scanptrlist.end(); scan++) 
    delete *scan;
}
